/* OmniVision OV9655 Camera Chip Support Code
 *
 * Copyright (c) 1999-2006 TomTom B.V <rogier.stam@tomtom.com> (08/08/2006) 
 * http://ovcam.org/ov511/
 *
 * Based upon the original ov7x10 driver by Mark McClelland <mark@ovcam.org>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */

/* This file is not a module yet */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include "ovcamchip_priv.h"
#include "ov9655.h"

/* Registers */
#define REG_BLUE		0x01	/* blue channel balance */
#define REG_RED			0x02	/* red channel balance */
#define REG_SAT			0x03	/* saturation */
#define REG_CNT			0x29	/* Y contrast */
#define REG_BRT			0x55	/* Y brightness */
#define REG_EXP1		0x04	/* manual exposure setting bit1:0 in bit 1:0 */
#define REG_EXP2		0x10	/* manual exposure setting bit9:2 in bit 7:0 */
#define REG_EXP3		0xA1	/* manual exposure setting bit15:10 in bit 5:0 */

static struct ovcamchip_regvals regvals_init_9xx0[] = {
{0xb5, 0x00},
{0x35, 0x00},
{0xa8, 0xc1},
{0x3a, 0xc2},   // Autobanding mode enabled.
{0xA3, 0x83},   // Default value for 1280x1024@15fps (max exposure line=1045), 50Hz Banding Filter value
{0xA2, 0x9D},   // Default value for 1280x1024@15fps (max exposure line=1045), 60Hz Banding Filter value
{0x13, 0xe7},   // Auto exposure control enabled.
{0x26, 0x72},
{0xab, 0x04},   // Reserved register ?
{0x6e, 0x00},   // Reserved register ?
{0x6d, 0x55},   // Reserved register ?
{0x10, 0x05},   // FOR SXGA Default setting. Not needed if init function sets default res.
{0xA1, 0x01},   //
{0x04, 0x01},   //
{0xbb, 0xae},   // Reserved register ?
{0x3e, 0x0c},   // Zoom off. Black & white pixel correction enabled.
{0x74, 0x10},	// 0x3a => Horizontal scale down coefficient
{0x75, 0x10},	// 0x35 => Vertical scale down coefficient 
{0x73, 0x00},	// 0x00 => Pixel clock output frequency adjustment.
{0xc7, 0x82},	// Pixel clock frequency selection.
{0x64, 0x02},   //
{0x65, 0x20},   //
{0x66, 0x01},   //
{0xc3, 0x4e},   // Reserved register ? 
{0xa4, 0x50},   // COM21. Partially reserved.
{0xaa, 0x92},   // Reserved register ?
{0xc2, 0x01},   // Reserved register ?
{0xc1, 0xc0},   // Reserved register ?
{0x1e, 0x14},   // Default VFLIP enabled.
{0xa9, 0xef},   // Reserved register ?	
{0x0e, 0x61},   // Exposure step can be set longer than VSYNC time. And reserved register.
{0x39, 0x57},   // AREF4.
{0x24, 0x3c},   // AGC/AEC stable operating region, upper limit.
{0x25, 0x36},   // AGC/AEC stable operating region, lower limit.
{0x12, 0x02},   // SXGA YUV mode.
{0x03, 0x02},	// VREF end low = 2 ipv 0.
{0x17, 0x18},	// 0x16 HREF start high 8 bit 16 ipv 18
{0x18, 0x04},	// 0x02 HREF stop high 8 bit 0x02 ipv 0x04
{0x1a, 0x81},	// 0x3d VREF stop high 8 bit 0x3d ipv 0x81
{0x36, 0x3a},	// 0xfa Array reference control. 0xFA ipv 0x3A.
{0x69, 0x0a},   // Use variopixel. Does this work with SXGA?
{0x8c, 0x8d},	// 0x8d UV average on/UV Adjust option=0x10 ipv off/0x80
{0xc0, 0xaa},   // Reserved register ?
{0x43, 0x0a},   // Reserved register ?
{0x44, 0xf0},   // Reserved register ?
{0x45, 0x46},   // Reserved register ?
{0x46, 0x62},   // Reserved register ?
{0x47, 0x2a},   // Reserved register ?
{0x48, 0x3c},   // Reserved register ?
{0x59, 0x85},   // AWB Control Option 1 t/m 6
{0x5a, 0xa9},   //
{0x5b, 0x64},   //
{0x5c, 0x84},   //
{0x5d, 0x53},   //
{0x5e, 0x0e},   //
{0x6c, 0x04},   // Reserved register ?
{0xc6, 0x85},   // Strobe enabled, histogram based.
{0xcc, 0xd8},   // Probability threshold for hrl to control aec/agc speed.
{0x71, 0x78},   // High reference luminance 
{0xa5, 0x68},   // Low reference luminance.
{0x6f, 0x9d},   // Reserved register ?
{0x42, 0xc0},   // Denoise strength auto adjust, Edge enhancement strength auto adjust.
{0x3f, 0x81},   // Edge enhancement factor.
{0x8a, 0x23},   // Reserved register ?
{0x14, 0x3a},   // Common control 9
{0x3b, 0xac}, 	// Nightmode: Auto Framerate Adjust enabled, min framerate=1/2
{0x34, 0x3d},   // Array reference control 1
{0x41, 0x40},   // Set reserved bit
{0xc9, 0xe0},   // Lower limit of probability for HRL after exposure/gain stabilizes.
{0xca, 0xe8},   // Upper limit of probability for HRL after exposure/gain stabilizes.
{0xcd, 0x93},   // Luminance threshold for AEC/AGC speed control
{0x7a, 0x20},   // Gamma curve highest segment slop.
{0x7b, 0x1c},   // Gamma curve segment input end point 0x10 - 0x340 segment 1 - 15.
{0x7c, 0x28},   //
{0x7d, 0x3c},   // 
{0x7e, 0x5a},   //
{0x7f, 0x68},   //
{0x80, 0x76},   //
{0x81, 0x80},   //
{0x82, 0x88},   //
{0x83, 0x8f},   //
{0x84, 0x96},   //
{0x85, 0xa3},   //
{0x86, 0xaf},   //
{0x87, 0xc4},   //
{0x88, 0xd7},   //
{0x89, 0xe8},   //
{0x4f, 0x98},   // Matrix coefficient 1 - 6
{0x50, 0x98},   //
{0x51, 0x00},   //
{0x52, 0x28},   //
{0x53, 0x70},   //
{0x54, 0x98},   // 
{0x58, 0x1a},   // Matrix coefficient sign register
{0x6b, 0x5a},   // Bypass internal regulator for DVDD // Should this be enabled?
{0x90, 0x7f},	// 0x86 Reserved ????
{0x91, 0x78},	// 0x84 Reserved ????
{0x9f, 0x83},	// 0x75 Gb Channel Blc value.
{0xa0, 0x7b},	// 0x73 Gr Channel Blc value.
{0x16, 0x24},   // Register 16 control. Auto add dummy frame if gain greater than 4x
{0x29, 0x15},   // RGB Pregain. One for R, G and B. Should this be higher?
{0x9d, 0x02},   // Lens Correction Option 6 & 7
{0x9e, 0x02},   //
{0xac, 0x80},   // BLC Control 1 - 8
{0xad, 0x80},   //
{0xae, 0x80},   //
{0xaf, 0x80},   //
{0xb2, 0xf2},   //
{0xb3, 0x20},   //
{0xb4, 0x20},   // Disable auto reference setting... Should this be changed?
{0x38, 0x12},   // ADC reference control.
{0x4a, 0xfa},	// 0xea Reserved ????
{0x4b, 0xe8},	// 0xe6 Reserved ????
{0x4c, 0x37},	// 0xe6 Reserved ????
{0x4d, 0x37},	// 0xe6 Reserved ????
{0x4e, 0x37},	// 0xe6 Reserved ????
{0xa6, 0x40},   // MAXIMUM AEC Banding step for 50Hz.
{0xbc, 0x7f},	// 0x04 ADC B channel offset setting
{0xbd, 0x7f},	// 0x01 ADC R channel offset setting
{0xbe, 0x7f},	// 0x03 ADC Gb channel offset setting
{0xbf, 0x7f},	// 0x01 ADC Gr channel offset setting
{0xff, 0xff},	/* END MARKER */
};

/* This initializes the OV9xx0 camera chip and relevant variables. */
static int ov9xx0_init(struct i2c_client *c)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	int rc;
	struct ov9655_private	*priv_dat=NULL;

	PDEBUG(4, "entered");
	rc = ov_write_regvals(c, regvals_init_9xx0);
	if (rc < 0)
		return rc;

	/* Private data. */
	ov->spriv = priv_dat = kmalloc(sizeof( *priv_dat ), GFP_KERNEL);
	if( ov->spriv == NULL )
		return -ENOMEM;

	/* Initialize. */
	memset( priv_dat, 0, sizeof( *priv_dat ) );
	priv_dat->image_properties.saturation=OV9XX0_DEF_SATURATION;
	priv_dat->image_properties.contrast=OV9XX0_DEF_CONTRAST;
	priv_dat->image_properties.brightness=OV9XX0_DEF_BRIGHTNESS;
	priv_dat->image_properties.hue=OV9XX0_DEF_HUE;
#ifdef OV_BUILD_HUESAT
	rc=ov9xx0_do_color_correct( c, priv_dat->image_properties.hue, priv_dat->image_properties.saturation );
	if( rc < 0 ) goto ov9xx0_init_error;
#endif
	rc=ov_write( c, 0x56, priv_dat->image_properties.contrast );
	if( rc < 0 ) goto ov9xx0_init_error;
	rc=ov_write( c, 0x55, priv_dat->image_properties.brightness );
	if( rc < 0 ) goto ov9xx0_init_error;

	/* Done. Set it as initialized. */
	ov->initialized=1;
	return rc;
ov9xx0_init_error:
	kfree( priv_dat );
	ov->spriv=NULL;
	return rc;
}

static int ov9xx0_free(struct i2c_client *c)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);

	/* Free the private data. */
	if( ov->spriv != NULL )
		kfree( ov->spriv );
	return 0;
}

static void ov9xx0_get_rate( struct i2c_client *c, unsigned char *rgbmod, unsigned short *base )
{
	unsigned char		regval=0;

	/* Get the current mode. */
	ov_read( c, 0x12, &regval );
	if( regval & 0x70 )
	{
		/* 30 fps mode (variopixel or not). */
		if( regval & 0x02 )
		{       
			/* Either RGB or YUV. */
			*rgbmod=1;
		}
		else
		{       
			/* One of the RAW RGB modes. */
			*rgbmod=3;
		}
                        
		*base=300;
	}
	else
	{       
		/* 15 fps mode. */
		if( regval & 0x02 )
		{       
			/* Either RGB or YUV. */
			*rgbmod=0;
		}
		else
		{       
			/* One of the RAW RGB modes. */
			*rgbmod=1;
		}
                       
		*base=150;
	}
	return;
}	

#ifdef OV_BUILD_HUESAT
static const int cos10k_array[]=
{10000,9999,9999,9999,9999,9999,9999,9999,9999,9998,9998,9998,9997,9997,9997,9996,9996,9995,9995,9994,
9993,9993,9992,9991,9991,9990,9989,9988,9988,9987,9986,9985,9984,9983,9982,9981,9980,9979,9978,9976,
9975,9974,9973,9971,9970,9969,9967,9966,9964,9963,9961,9960,9958,9957,9955,9953,9952,9950,9948,9947,
9945,9943,9941,9939,9937,9935,9933,9931,9929,9927,9925,9923,9921,9918,9916,9914,9912,9909,9907,9905,
9902,9900,9897,9895,9892,9890,9887,9884,9882,9879,9876,9874,9871,9868,9865,9862,9859,9857,9854,9851,
9848,9845,9841,9838,9835,9832,9829,9826,9822,9819,9816,9812,9809,9806,9802,9799,9795,9792,9788,9785,
9781,9777,9774,9770,9766,9762,9759,9755,9751,9747,9743,9739,9735,9731,9727,9723,9719,9715,9711,9707,
9702,9698,9694,9690,9685,9681,9677,9672,9668,9663,9659,9654,9650,9645,9640,9636,9631,9626,9622,9617,
9612,9607,9602,9598,9593,9588,9583,9578,9573,9568,9563,9557,9552,9547,9542,9537,9531,9526,9521,9515,
9510,9505,9499,9494,9488,9483,9477,9472,9466,9460,9455,9449,9443,9438,9432,9426,9420,9414,9408,9402,
9396,9390,9384,9378,9372,9366,9360,9354,9348,9342,9335,9329,9323,9316,9310,9304,9297,9291,9284,9278,
9271,9265,9258,9252,9245,9238,9232,9225,9218,9211,9205,9198,9191,9184,9177,9170,9163,9156,9149,9142,
9135,9128,9121,9114,9106,9099,9092,9085,9077,9070,9063,9055,9048,9040,9033,9025,9018,9010,9003,8995,
8987,8980,8972,8964,8957,8949,8941,8933,8925,8917,8910,8902,8894,8886,8878,8870,8862,8853,8845,8837,
8829,8821,8813,8804,8796,8788,8779,8771,8763,8754,8746,8737,8729,8720,8712,8703,8694,8686,8677,8668,
8660,8651,8642,8633,8625,8616,8607,8598,8589,8580,8571,8562,8553,8544,8535,8526,8517,8508,8498,8489,
8480,8471,8461,8452,8443,8433,8424,8415,8405,8396,8386,8377,8367,8358,8348,8338,8329,8319,8309,8300,
8290,8280,8270,8260,8251,8241,8231,8221,8211,8201,8191,8181,8171,8161,8151,8141,8131,8120,8110,8100,
8090,8079,8069,8059,8048,8038,8028,8017,8007,7996,7986,7975,7965,7954,7944,7933,7922,7912,7901,7890,
7880,7869,7858,7847,7836,7826,7815,7804,7793,7782,7771,7760,7749,7738,7727,7716,7705,7693,7682,7671,
7660,7649,7637,7626,7615,7604,7592,7581,7569,7558,7547,7535,7524,7512,7501,7489,7477,7466,7454,7443,
7431,7419,7408,7396,7384,7372,7360,7349,7337,7325,7313,7301,7289,7277,7265,7253,7241,7229,7217,7205,
7193,7181,7169,7156,7144,7132,7120,7107,7095,7083,7071,7058,7046,7033,7021,7009,6996,6984,6971,6959,
6946,6934,6921,6908,6896,6883,6870,6858,6845,6832,6819,6807,6794,6781,6768,6755,6743,6730,6717,6704,
6691,6678,6665,6652,6639,6626,6613,6600,6586,6573,6560,6547,6534,6520,6507,6494,6481,6467,6454,6441,
6427,6414,6401,6387,6374,6360,6347,6333,6320,6306,6293,6279,6266,6252,6238,6225,6211,6197,6184,6170,
6156,6142,6129,6115,6101,6087,6073,6059,6045,6032,6018,6004,5990,5976,5962,5948,5934,5920,5906,5891,
5877,5863,5849,5835,5821,5807,5792,5778,5764,5750,5735,5721,5707,5692,5678,5664,5649,5635,5620,5606,
5591,5577,5562,5548,5533,5519,5504,5490,5475,5461,5446,5431,5417,5402,5387,5372,5358,5343,5328,5313,
5299,5284,5269,5254,5239,5224,5210,5195,5180,5165,5150,5135,5120,5105,5090,5075,5060,5045,5030,5015,
5000,4984,4969,4954,4939,4924,4909,4893,4878,4863,4848,4832,4817,4802,4786,4771,4756,4740,4725,4710,
4694,4679,4663,4648,4632,4617,4601,4586,4570,4555,4539,4524,4508,4493,4477,4461,4446,4430,4415,4399,
4383,4368,4352,4336,4320,4305,4289,4273,4257,4241,4226,4210,4194,4178,4162,4146,4131,4115,4099,4083,
4067,4051,4035,4019,4003,3987,3971,3955,3939,3923,3907,3891,3875,3859,3842,3826,3810,3794,3778,3762,
3746,3729,3713,3697,3681,3665,3648,3632,3616,3599,3583,3567,3551,3534,3518,3502,3485,3469,3452,3436,
3420,3403,3387,3370,3354,3338,3321,3305,3288,3272,3255,3239,3222,3206,3189,3173,3156,3139,3123,3106,
3090,3073,3056,3040,3023,3007,2990,2973,2957,2940,2923,2907,2890,2873,2856,2840,2823,2806,2789,2773,
2756,2739,2722,2706,2689,2672,2655,2638,2621,2605,2588,2571,2554,2537,2520,2503,2486,2469,2453,2436,
2419,2402,2385,2368,2351,2334,2317,2300,2283,2266,2249,2232,2215,2198,2181,2164,2147,2130,2113,2096,
2079,2062,2044,2027,2010,1993,1976,1959,1942,1925,1908,1890,1873,1856,1839,1822,1805,1788,1770,1753,
1736,1719,1702,1684,1667,1650,1633,1616,1598,1581,1564,1547,1529,1512,1495,1478,1460,1443,1426,1409,
1391,1374,1357,1339,1322,1305,1287,1270,1253,1236,1218,1201,1184,1166,1149,1132,1114,1097,1079,1062,
1045,1027,1010,993,975,958,941,923,906,888,871,854,836,819,801,784,767,749,732,714,
697,680,662,645,627,610,593,575,558,540,523,505,488,471,453,436,418,401,383,366,
348,331,314,296,279,261,244,226,209,191,174,157,139,122,104,87,69,52,34,17,
0,-17,-34,-52,-69,-87,-104,-122,-139,-157,-174,-191,-209,-226,-244,-261,-279,-296,-314,-331,
-348,-366,-383,-401,-418,-436,-453,-471,-488,-505,-523,-540,-558,-575,-593,-610,-627,-645,-662,-680,
-697,-714,-732,-749,-767,-784,-801,-819,-836,-854,-871,-888,-906,-923,-941,-958,-975,-993,-1010,-1027,
-1045,-1062,-1079,-1097,-1114,-1132,-1149,-1166,-1184,-1201,-1218,-1236,-1253,-1270,-1287,-1305,-1322,-1339,-1357,-1374,
-1391,-1409,-1426,-1443,-1460,-1478,-1495,-1512,-1529,-1547,-1564,-1581,-1598,-1616,-1633,-1650,-1667,-1684,-1702,-1719,
-1736,-1753,-1770,-1788,-1805,-1822,-1839,-1856,-1873,-1890,-1908,-1925,-1942,-1959,-1976,-1993,-2010,-2027,-2044,-2062,
-2079,-2096,-2113,-2130,-2147,-2164,-2181,-2198,-2215,-2232,-2249,-2266,-2283,-2300,-2317,-2334,-2351,-2368,-2385,-2402,
-2419,-2436,-2453,-2469,-2486,-2503,-2520,-2537,-2554,-2571,-2588,-2605,-2621,-2638,-2655,-2672,-2689,-2706,-2722,-2739,
-2756,-2773,-2789,-2806,-2823,-2840,-2856,-2873,-2890,-2907,-2923,-2940,-2957,-2973,-2990,-3007,-3023,-3040,-3056,-3073,
-3090,-3106,-3123,-3139,-3156,-3173,-3189,-3206,-3222,-3239,-3255,-3272,-3288,-3305,-3321,-3338,-3354,-3370,-3387,-3403,
-3420,-3436,-3452,-3469,-3485,-3502,-3518,-3534,-3551,-3567,-3583,-3599,-3616,-3632,-3648,-3665,-3681,-3697,-3713,-3729,
-3746,-3762,-3778,-3794,-3810,-3826,-3842,-3859,-3875,-3891,-3907,-3923,-3939,-3955,-3971,-3987,-4003,-4019,-4035,-4051,
-4067,-4083,-4099,-4115,-4131,-4146,-4162,-4178,-4194,-4210,-4226,-4241,-4257,-4273,-4289,-4305,-4320,-4336,-4352,-4368,
-4383,-4399,-4415,-4430,-4446,-4461,-4477,-4493,-4508,-4524,-4539,-4555,-4570,-4586,-4601,-4617,-4632,-4648,-4663,-4679,
-4694,-4710,-4725,-4740,-4756,-4771,-4786,-4802,-4817,-4832,-4848,-4863,-4878,-4893,-4909,-4924,-4939,-4954,-4969,-4984,
-4999,-5015,-5030,-5045,-5060,-5075,-5090,-5105,-5120,-5135,-5150,-5165,-5180,-5195,-5210,-5224,-5239,-5254,-5269,-5284,
-5299,-5313,-5328,-5343,-5358,-5372,-5387,-5402,-5417,-5431,-5446,-5461,-5475,-5490,-5504,-5519,-5533,-5548,-5562,-5577,
-5591,-5606,-5620,-5635,-5649,-5664,-5678,-5692,-5707,-5721,-5735,-5750,-5764,-5778,-5792,-5807,-5821,-5835,-5849,-5863,
-5877,-5891,-5906,-5920,-5934,-5948,-5962,-5976,-5990,-6004,-6018,-6032,-6045,-6059,-6073,-6087,-6101,-6115,-6129,-6142,
-6156,-6170,-6184,-6197,-6211,-6225,-6238,-6252,-6266,-6279,-6293,-6306,-6320,-6333,-6347,-6360,-6374,-6387,-6401,-6414,
-6427,-6441,-6454,-6467,-6481,-6494,-6507,-6520,-6534,-6547,-6560,-6573,-6586,-6600,-6613,-6626,-6639,-6652,-6665,-6678,
-6691,-6704,-6717,-6730,-6743,-6755,-6768,-6781,-6794,-6807,-6819,-6832,-6845,-6858,-6870,-6883,-6896,-6908,-6921,-6934,
-6946,-6959,-6971,-6984,-6996,-7009,-7021,-7033,-7046,-7058,-7071,-7083,-7095,-7107,-7120,-7132,-7144,-7156,-7169,-7181,
-7193,-7205,-7217,-7229,-7241,-7253,-7265,-7277,-7289,-7301,-7313,-7325,-7337,-7349,-7360,-7372,-7384,-7396,-7408,-7419,
-7431,-7443,-7454,-7466,-7477,-7489,-7501,-7512,-7524,-7535,-7547,-7558,-7569,-7581,-7592,-7604,-7615,-7626,-7637,-7649,
-7660,-7671,-7682,-7693,-7705,-7716,-7727,-7738,-7749,-7760,-7771,-7782,-7793,-7804,-7815,-7826,-7836,-7847,-7858,-7869,
-7880,-7890,-7901,-7912,-7922,-7933,-7944,-7954,-7965,-7975,-7986,-7996,-8007,-8017,-8028,-8038,-8048,-8059,-8069,-8079,
-8090,-8100,-8110,-8120,-8131,-8141,-8151,-8161,-8171,-8181,-8191,-8201,-8211,-8221,-8231,-8241,-8251,-8260,-8270,-8280,
-8290,-8300,-8309,-8319,-8329,-8338,-8348,-8358,-8367,-8377,-8386,-8396,-8405,-8415,-8424,-8433,-8443,-8452,-8461,-8471,
-8480,-8489,-8498,-8508,-8517,-8526,-8535,-8544,-8553,-8562,-8571,-8580,-8589,-8598,-8607,-8616,-8625,-8633,-8642,-8651,
-8660,-8668,-8677,-8686,-8694,-8703,-8712,-8720,-8729,-8737,-8746,-8754,-8763,-8771,-8779,-8788,-8796,-8804,-8813,-8821,
-8829,-8837,-8845,-8853,-8862,-8870,-8878,-8886,-8894,-8902,-8910,-8917,-8925,-8933,-8941,-8949,-8957,-8964,-8972,-8980,
-8987,-8995,-9003,-9010,-9018,-9025,-9033,-9040,-9048,-9055,-9063,-9070,-9077,-9085,-9092,-9099,-9106,-9114,-9121,-9128,
-9135,-9142,-9149,-9156,-9163,-9170,-9177,-9184,-9191,-9198,-9205,-9211,-9218,-9225,-9232,-9238,-9245,-9252,-9258,-9265,
-9271,-9278,-9284,-9291,-9297,-9304,-9310,-9316,-9323,-9329,-9335,-9342,-9348,-9354,-9360,-9366,-9372,-9378,-9384,-9390,
-9396,-9402,-9408,-9414,-9420,-9426,-9432,-9438,-9443,-9449,-9455,-9460,-9466,-9472,-9477,-9483,-9488,-9494,-9499,-9505,
-9510,-9515,-9521,-9526,-9531,-9537,-9542,-9547,-9552,-9557,-9563,-9568,-9573,-9578,-9583,-9588,-9593,-9598,-9602,-9607,
-9612,-9617,-9622,-9626,-9631,-9636,-9640,-9645,-9650,-9654,-9659,-9663,-9668,-9672,-9677,-9681,-9685,-9690,-9694,-9698,
-9702,-9707,-9711,-9715,-9719,-9723,-9727,-9731,-9735,-9739,-9743,-9747,-9751,-9755,-9759,-9762,-9766,-9770,-9774,-9777,
-9781,-9785,-9788,-9792,-9795,-9799,-9802,-9806,-9809,-9812,-9816,-9819,-9822,-9826,-9829,-9832,-9835,-9838,-9841,-9845,
-9848,-9851,-9854,-9857,-9859,-9862,-9865,-9868,-9871,-9874,-9876,-9879,-9882,-9884,-9887,-9890,-9892,-9895,-9897,-9900,
-9902,-9905,-9907,-9909,-9912,-9914,-9916,-9918,-9921,-9923,-9925,-9927,-9929,-9931,-9933,-9935,-9937,-9939,-9941,-9943,
-9945,-9947,-9948,-9950,-9952,-9953,-9955,-9957,-9958,-9960,-9961,-9963,-9964,-9966,-9967,-9969,-9970,-9971,-9973,-9974,
-9975,-9976,-9978,-9979,-9980,-9981,-9982,-9983,-9984,-9985,-9986,-9987,-9988,-9988,-9989,-9990,-9991,-9991,-9992,-9993,
-9993,-9994,-9995,-9995,-9996,-9996,-9997,-9997,-9997,-9998,-9998,-9998,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9999,
-10000,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9998,-9998,-9998,-9997,-9997,-9997,-9996,-9996,-9995,-9995,-9994,
-9993,-9993,-9992,-9991,-9991,-9990,-9989,-9988,-9988,-9987,-9986,-9985,-9984,-9983,-9982,-9981,-9980,-9979,-9978,-9976,
-9975,-9974,-9973,-9971,-9970,-9969,-9967,-9966,-9964,-9963,-9961,-9960,-9958,-9957,-9955,-9953,-9952,-9950,-9948,-9947,
-9945,-9943,-9941,-9939,-9937,-9935,-9933,-9931,-9929,-9927,-9925,-9923,-9921,-9918,-9916,-9914,-9912,-9909,-9907,-9905,
-9902,-9900,-9897,-9895,-9892,-9890,-9887,-9884,-9882,-9879,-9876,-9874,-9871,-9868,-9865,-9862,-9859,-9857,-9854,-9851,
-9848,-9845,-9841,-9838,-9835,-9832,-9829,-9826,-9822,-9819,-9816,-9812,-9809,-9806,-9802,-9799,-9795,-9792,-9788,-9785,
-9781,-9777,-9774,-9770,-9766,-9762,-9759,-9755,-9751,-9747,-9743,-9739,-9735,-9731,-9727,-9723,-9719,-9715,-9711,-9707,
-9702,-9698,-9694,-9690,-9685,-9681,-9677,-9672,-9668,-9663,-9659,-9654,-9650,-9645,-9640,-9636,-9631,-9626,-9622,-9617,
-9612,-9607,-9602,-9598,-9593,-9588,-9583,-9578,-9573,-9568,-9563,-9557,-9552,-9547,-9542,-9537,-9531,-9526,-9521,-9515,
-9510,-9505,-9499,-9494,-9488,-9483,-9477,-9472,-9466,-9460,-9455,-9449,-9443,-9438,-9432,-9426,-9420,-9414,-9408,-9402,
-9396,-9390,-9384,-9378,-9372,-9366,-9360,-9354,-9348,-9342,-9335,-9329,-9323,-9316,-9310,-9304,-9297,-9291,-9284,-9278,
-9271,-9265,-9258,-9252,-9245,-9238,-9232,-9225,-9218,-9211,-9205,-9198,-9191,-9184,-9177,-9170,-9163,-9156,-9149,-9142,
-9135,-9128,-9121,-9114,-9106,-9099,-9092,-9085,-9077,-9070,-9063,-9055,-9048,-9040,-9033,-9025,-9018,-9010,-9003,-8995,
-8987,-8980,-8972,-8964,-8957,-8949,-8941,-8933,-8925,-8917,-8910,-8902,-8894,-8886,-8878,-8870,-8862,-8853,-8845,-8837,
-8829,-8821,-8813,-8804,-8796,-8788,-8779,-8771,-8763,-8754,-8746,-8737,-8729,-8720,-8712,-8703,-8694,-8686,-8677,-8668,
-8660,-8651,-8642,-8633,-8625,-8616,-8607,-8598,-8589,-8580,-8571,-8562,-8553,-8544,-8535,-8526,-8517,-8508,-8498,-8489,
-8480,-8471,-8461,-8452,-8443,-8433,-8424,-8415,-8405,-8396,-8386,-8377,-8367,-8358,-8348,-8338,-8329,-8319,-8309,-8300,
-8290,-8280,-8270,-8260,-8251,-8241,-8231,-8221,-8211,-8201,-8191,-8181,-8171,-8161,-8151,-8141,-8131,-8120,-8110,-8100,
-8090,-8079,-8069,-8059,-8048,-8038,-8028,-8017,-8007,-7996,-7986,-7975,-7965,-7954,-7944,-7933,-7922,-7912,-7901,-7890,
-7880,-7869,-7858,-7847,-7836,-7826,-7815,-7804,-7793,-7782,-7771,-7760,-7749,-7738,-7727,-7716,-7705,-7693,-7682,-7671,
-7660,-7649,-7637,-7626,-7615,-7604,-7592,-7581,-7569,-7558,-7547,-7535,-7524,-7512,-7501,-7489,-7477,-7466,-7454,-7443,
-7431,-7419,-7408,-7396,-7384,-7372,-7360,-7349,-7337,-7325,-7313,-7301,-7289,-7277,-7265,-7253,-7241,-7229,-7217,-7205,
-7193,-7181,-7169,-7156,-7144,-7132,-7120,-7107,-7095,-7083,-7071,-7058,-7046,-7033,-7021,-7009,-6996,-6984,-6971,-6959,
-6946,-6934,-6921,-6908,-6896,-6883,-6870,-6858,-6845,-6832,-6819,-6807,-6794,-6781,-6768,-6755,-6743,-6730,-6717,-6704,
-6691,-6678,-6665,-6652,-6639,-6626,-6613,-6600,-6586,-6573,-6560,-6547,-6534,-6520,-6507,-6494,-6481,-6467,-6454,-6441,
-6427,-6414,-6401,-6387,-6374,-6360,-6347,-6333,-6320,-6306,-6293,-6279,-6266,-6252,-6238,-6225,-6211,-6197,-6184,-6170,
-6156,-6142,-6129,-6115,-6101,-6087,-6073,-6059,-6045,-6032,-6018,-6004,-5990,-5976,-5962,-5948,-5934,-5920,-5906,-5891,
-5877,-5863,-5849,-5835,-5821,-5807,-5792,-5778,-5764,-5750,-5735,-5721,-5707,-5692,-5678,-5664,-5649,-5635,-5620,-5606,
-5591,-5577,-5562,-5548,-5533,-5519,-5504,-5490,-5475,-5461,-5446,-5431,-5417,-5402,-5387,-5372,-5358,-5343,-5328,-5313,
-5299,-5284,-5269,-5254,-5239,-5224,-5210,-5195,-5180,-5165,-5150,-5135,-5120,-5105,-5090,-5075,-5060,-5045,-5030,-5015,
-5000,-4984,-4969,-4954,-4939,-4924,-4909,-4893,-4878,-4863,-4848,-4832,-4817,-4802,-4786,-4771,-4756,-4740,-4725,-4710,
-4694,-4679,-4663,-4648,-4632,-4617,-4601,-4586,-4570,-4555,-4539,-4524,-4508,-4493,-4477,-4461,-4446,-4430,-4415,-4399,
-4383,-4368,-4352,-4336,-4320,-4305,-4289,-4273,-4257,-4241,-4226,-4210,-4194,-4178,-4162,-4146,-4131,-4115,-4099,-4083,
-4067,-4051,-4035,-4019,-4003,-3987,-3971,-3955,-3939,-3923,-3907,-3891,-3875,-3859,-3842,-3826,-3810,-3794,-3778,-3762,
-3746,-3729,-3713,-3697,-3681,-3665,-3648,-3632,-3616,-3599,-3583,-3567,-3551,-3534,-3518,-3502,-3485,-3469,-3452,-3436,
-3420,-3403,-3387,-3370,-3354,-3338,-3321,-3305,-3288,-3272,-3255,-3239,-3222,-3206,-3189,-3173,-3156,-3139,-3123,-3106,
-3090,-3073,-3056,-3040,-3023,-3007,-2990,-2973,-2957,-2940,-2923,-2907,-2890,-2873,-2856,-2840,-2823,-2806,-2789,-2773,
-2756,-2739,-2722,-2706,-2689,-2672,-2655,-2638,-2621,-2605,-2588,-2571,-2554,-2537,-2520,-2503,-2486,-2469,-2453,-2436,
-2419,-2402,-2385,-2368,-2351,-2334,-2317,-2300,-2283,-2266,-2249,-2232,-2215,-2198,-2181,-2164,-2147,-2130,-2113,-2096,
-2079,-2062,-2044,-2027,-2010,-1993,-1976,-1959,-1942,-1925,-1908,-1890,-1873,-1856,-1839,-1822,-1805,-1788,-1770,-1753,
-1736,-1719,-1702,-1684,-1667,-1650,-1633,-1616,-1598,-1581,-1564,-1547,-1529,-1512,-1495,-1478,-1460,-1443,-1426,-1409,
-1391,-1374,-1357,-1339,-1322,-1305,-1287,-1270,-1253,-1236,-1218,-1201,-1184,-1166,-1149,-1132,-1114,-1097,-1079,-1062,
-1045,-1027,-1010,-993,-975,-958,-941,-923,-906,-888,-871,-854,-836,-819,-801,-784,-767,-749,-732,-714,
-697,-680,-662,-645,-627,-610,-593,-575,-558,-540,-523,-505,-488,-471,-453,-436,-418,-401,-383,-366,
-348,-331,-314,-296,-279,-261,-244,-226,-209,-191,-174,-157,-139,-122,-104,-87,-69,-52,-34,-17,
0,17,34,52,69,87,104,122,139,157,174,191,209,226,244,261,279,296,314,331,
348,366,383,401,418,436,453,471,488,505,523,540,558,575,593,610,627,645,662,680,
697,714,732,749,767,784,801,819,836,854,871,888,906,923,941,958,975,993,1010,1027,
1045,1062,1079,1097,1114,1132,1149,1166,1184,1201,1218,1236,1253,1270,1287,1305,1322,1339,1357,1374,
1391,1409,1426,1443,1460,1478,1495,1512,1529,1547,1564,1581,1598,1616,1633,1650,1667,1684,1702,1719,
1736,1753,1770,1788,1805,1822,1839,1856,1873,1890,1908,1925,1942,1959,1976,1993,2010,2027,2044,2062,
2079,2096,2113,2130,2147,2164,2181,2198,2215,2232,2249,2266,2283,2300,2317,2334,2351,2368,2385,2402,
2419,2436,2453,2469,2486,2503,2520,2537,2554,2571,2588,2605,2621,2638,2655,2672,2689,2706,2722,2739,
2756,2773,2789,2806,2823,2840,2856,2873,2890,2907,2923,2940,2957,2973,2990,3007,3023,3040,3056,3073,
3090,3106,3123,3139,3156,3173,3189,3206,3222,3239,3255,3272,3288,3305,3321,3338,3354,3370,3387,3403,
3420,3436,3452,3469,3485,3502,3518,3534,3551,3567,3583,3599,3616,3632,3648,3665,3681,3697,3713,3729,
3746,3762,3778,3794,3810,3826,3842,3859,3875,3891,3907,3923,3939,3955,3971,3987,4003,4019,4035,4051,
4067,4083,4099,4115,4131,4146,4162,4178,4194,4210,4226,4241,4257,4273,4289,4305,4320,4336,4352,4368,
4383,4399,4415,4430,4446,4461,4477,4493,4508,4524,4539,4555,4570,4586,4601,4617,4632,4648,4663,4679,
4694,4710,4725,4740,4756,4771,4786,4802,4817,4832,4848,4863,4878,4893,4909,4924,4939,4954,4969,4984,
5000,5015,5030,5045,5060,5075,5090,5105,5120,5135,5150,5165,5180,5195,5210,5224,5239,5254,5269,5284,
5299,5313,5328,5343,5358,5372,5387,5402,5417,5431,5446,5461,5475,5490,5504,5519,5533,5548,5562,5577,
5591,5606,5620,5635,5649,5664,5678,5692,5707,5721,5735,5750,5764,5778,5792,5807,5821,5835,5849,5863,
5877,5891,5906,5920,5934,5948,5962,5976,5990,6004,6018,6032,6045,6059,6073,6087,6101,6115,6129,6142,
6156,6170,6184,6197,6211,6225,6238,6252,6266,6279,6293,6306,6320,6333,6347,6360,6374,6387,6401,6414,
6427,6441,6454,6467,6481,6494,6507,6520,6534,6547,6560,6573,6586,6600,6613,6626,6639,6652,6665,6678,
6691,6704,6717,6730,6743,6755,6768,6781,6794,6807,6819,6832,6845,6858,6870,6883,6896,6908,6921,6934,
6946,6959,6971,6984,6996,7009,7021,7033,7046,7058,7071,7083,7095,7107,7120,7132,7144,7156,7169,7181,
7193,7205,7217,7229,7241,7253,7265,7277,7289,7301,7313,7325,7337,7349,7360,7372,7384,7396,7408,7419,
7431,7443,7454,7466,7477,7489,7501,7512,7524,7535,7547,7558,7569,7581,7592,7604,7615,7626,7637,7649,
7660,7671,7682,7693,7705,7716,7727,7738,7749,7760,7771,7782,7793,7804,7815,7826,7836,7847,7858,7869,
7880,7890,7901,7912,7922,7933,7944,7954,7965,7975,7986,7996,8007,8017,8028,8038,8048,8059,8069,8079,
8090,8100,8110,8120,8131,8141,8151,8161,8171,8181,8191,8201,8211,8221,8231,8241,8251,8260,8270,8280,
8290,8300,8309,8319,8329,8338,8348,8358,8367,8377,8386,8396,8405,8415,8424,8433,8443,8452,8461,8471,
8480,8489,8498,8508,8517,8526,8535,8544,8553,8562,8571,8580,8589,8598,8607,8616,8625,8633,8642,8651,
8660,8668,8677,8686,8694,8703,8712,8720,8729,8737,8746,8754,8763,8771,8779,8788,8796,8804,8813,8821,
8829,8837,8845,8853,8862,8870,8878,8886,8894,8902,8910,8917,8925,8933,8941,8949,8957,8964,8972,8980,
8987,8995,9003,9010,9018,9025,9033,9040,9048,9055,9063,9070,9077,9085,9092,9099,9106,9114,9121,9128,
9135,9142,9149,9156,9163,9170,9177,9184,9191,9198,9205,9211,9218,9225,9232,9238,9245,9252,9258,9265,
9271,9278,9284,9291,9297,9304,9310,9316,9323,9329,9335,9342,9348,9354,9360,9366,9372,9378,9384,9390,
9396,9402,9408,9414,9420,9426,9432,9438,9443,9449,9455,9460,9466,9472,9477,9483,9488,9494,9499,9505,
9510,9515,9521,9526,9531,9537,9542,9547,9552,9557,9563,9568,9573,9578,9583,9588,9593,9598,9602,9607,
9612,9617,9622,9626,9631,9636,9640,9645,9650,9654,9659,9663,9668,9672,9677,9681,9685,9690,9694,9698,
9702,9707,9711,9715,9719,9723,9727,9731,9735,9739,9743,9747,9751,9755,9759,9762,9766,9770,9774,9777,
9781,9785,9788,9792,9795,9799,9802,9806,9809,9812,9816,9819,9822,9826,9829,9832,9835,9838,9841,9845,
9848,9851,9854,9857,9859,9862,9865,9868,9871,9874,9876,9879,9882,9884,9887,9890,9892,9895,9897,9900,
9902,9905,9907,9909,9912,9914,9916,9918,9921,9923,9925,9927,9929,9931,9933,9935,9937,9939,9941,9943,
9945,9947,9948,9950,9952,9953,9955,9957,9958,9960,9961,9963,9964,9966,9967,9969,9970,9971,9973,9974,
9975,9976,9978,9979,9980,9981,9982,9983,9984,9985,9986,9987,9988,9988,9989,9990,9991,9991,9992,9993,
9993,9994,9995,9995,9996,9996,9997,9997,9997,9998,9998,9998,9999,9999,9999,9999,9999,9999,9999,9999
};

static const int sin10k_array[]=
{0,17,34,52,69,87,104,122,139,157,174,191,209,226,244,261,279,296,314,331,
348,366,383,401,418,436,453,471,488,505,523,540,558,575,593,610,627,645,662,680,
697,714,732,749,767,784,801,819,836,854,871,888,906,923,941,958,975,993,1010,1027,
1045,1062,1079,1097,1114,1132,1149,1166,1184,1201,1218,1236,1253,1270,1287,1305,1322,1339,1357,1374,
1391,1409,1426,1443,1460,1478,1495,1512,1529,1547,1564,1581,1598,1616,1633,1650,1667,1684,1702,1719,
1736,1753,1770,1788,1805,1822,1839,1856,1873,1890,1908,1925,1942,1959,1976,1993,2010,2027,2044,2062,
2079,2096,2113,2130,2147,2164,2181,2198,2215,2232,2249,2266,2283,2300,2317,2334,2351,2368,2385,2402,
2419,2436,2453,2469,2486,2503,2520,2537,2554,2571,2588,2605,2621,2638,2655,2672,2689,2706,2722,2739,
2756,2773,2789,2806,2823,2840,2856,2873,2890,2907,2923,2940,2957,2973,2990,3007,3023,3040,3056,3073,
3090,3106,3123,3139,3156,3173,3189,3206,3222,3239,3255,3272,3288,3305,3321,3338,3354,3370,3387,3403,
3420,3436,3452,3469,3485,3502,3518,3534,3551,3567,3583,3599,3616,3632,3648,3665,3681,3697,3713,3729,
3746,3762,3778,3794,3810,3826,3842,3859,3875,3891,3907,3923,3939,3955,3971,3987,4003,4019,4035,4051,
4067,4083,4099,4115,4131,4146,4162,4178,4194,4210,4226,4241,4257,4273,4289,4305,4320,4336,4352,4368,
4383,4399,4415,4430,4446,4461,4477,4493,4508,4524,4539,4555,4570,4586,4601,4617,4632,4648,4663,4679,
4694,4710,4725,4740,4756,4771,4786,4802,4817,4832,4848,4863,4878,4893,4909,4924,4939,4954,4969,4984,
4999,5015,5030,5045,5060,5075,5090,5105,5120,5135,5150,5165,5180,5195,5210,5224,5239,5254,5269,5284,
5299,5313,5328,5343,5358,5372,5387,5402,5417,5431,5446,5461,5475,5490,5504,5519,5533,5548,5562,5577,
5591,5606,5620,5635,5649,5664,5678,5692,5707,5721,5735,5750,5764,5778,5792,5807,5821,5835,5849,5863,
5877,5891,5906,5920,5934,5948,5962,5976,5990,6004,6018,6032,6045,6059,6073,6087,6101,6115,6129,6142,
6156,6170,6184,6197,6211,6225,6238,6252,6266,6279,6293,6306,6320,6333,6347,6360,6374,6387,6401,6414,
6427,6441,6454,6467,6481,6494,6507,6520,6534,6547,6560,6573,6586,6600,6613,6626,6639,6652,6665,6678,
6691,6704,6717,6730,6743,6755,6768,6781,6794,6807,6819,6832,6845,6858,6870,6883,6896,6908,6921,6934,
6946,6959,6971,6984,6996,7009,7021,7033,7046,7058,7071,7083,7095,7107,7120,7132,7144,7156,7169,7181,
7193,7205,7217,7229,7241,7253,7265,7277,7289,7301,7313,7325,7337,7349,7360,7372,7384,7396,7408,7419,
7431,7443,7454,7466,7477,7489,7501,7512,7524,7535,7547,7558,7569,7581,7592,7604,7615,7626,7637,7649,
7660,7671,7682,7693,7705,7716,7727,7738,7749,7760,7771,7782,7793,7804,7815,7826,7836,7847,7858,7869,
7880,7890,7901,7912,7922,7933,7944,7954,7965,7975,7986,7996,8007,8017,8028,8038,8048,8059,8069,8079,
8090,8100,8110,8120,8131,8141,8151,8161,8171,8181,8191,8201,8211,8221,8231,8241,8251,8260,8270,8280,
8290,8300,8309,8319,8329,8338,8348,8358,8367,8377,8386,8396,8405,8415,8424,8433,8443,8452,8461,8471,
8480,8489,8498,8508,8517,8526,8535,8544,8553,8562,8571,8580,8589,8598,8607,8616,8625,8633,8642,8651,
8660,8668,8677,8686,8694,8703,8712,8720,8729,8737,8746,8754,8763,8771,8779,8788,8796,8804,8813,8821,
8829,8837,8845,8853,8862,8870,8878,8886,8894,8902,8910,8917,8925,8933,8941,8949,8957,8964,8972,8980,
8987,8995,9003,9010,9018,9025,9033,9040,9048,9055,9063,9070,9077,9085,9092,9099,9106,9114,9121,9128,
9135,9142,9149,9156,9163,9170,9177,9184,9191,9198,9205,9211,9218,9225,9232,9238,9245,9252,9258,9265,
9271,9278,9284,9291,9297,9304,9310,9316,9323,9329,9335,9342,9348,9354,9360,9366,9372,9378,9384,9390,
9396,9402,9408,9414,9420,9426,9432,9438,9443,9449,9455,9460,9466,9472,9477,9483,9488,9494,9499,9505,
9510,9515,9521,9526,9531,9537,9542,9547,9552,9557,9563,9568,9573,9578,9583,9588,9593,9598,9602,9607,
9612,9617,9622,9626,9631,9636,9640,9645,9650,9654,9659,9663,9668,9672,9677,9681,9685,9690,9694,9698,
9702,9707,9711,9715,9719,9723,9727,9731,9735,9739,9743,9747,9751,9755,9759,9762,9766,9770,9774,9777,
9781,9785,9788,9792,9795,9799,9802,9806,9809,9812,9816,9819,9822,9826,9829,9832,9835,9838,9841,9845,
9848,9851,9854,9857,9859,9862,9865,9868,9871,9874,9876,9879,9882,9884,9887,9890,9892,9895,9897,9900,
9902,9905,9907,9909,9912,9914,9916,9918,9921,9923,9925,9927,9929,9931,9933,9935,9937,9939,9941,9943,
9945,9947,9948,9950,9952,9953,9955,9957,9958,9960,9961,9963,9964,9966,9967,9969,9970,9971,9973,9974,
9975,9976,9978,9979,9980,9981,9982,9983,9984,9985,9986,9987,9988,9988,9989,9990,9991,9991,9992,9993,
9993,9994,9995,9995,9996,9996,9997,9997,9997,9998,9998,9998,9999,9999,9999,9999,9999,9999,9999,9999,
10000,9999,9999,9999,9999,9999,9999,9999,9999,9998,9998,9998,9997,9997,9997,9996,9996,9995,9995,9994,
9993,9993,9992,9991,9991,9990,9989,9988,9988,9987,9986,9985,9984,9983,9982,9981,9980,9979,9978,9976,
9975,9974,9973,9971,9970,9969,9967,9966,9964,9963,9961,9960,9958,9957,9955,9953,9952,9950,9948,9947,
9945,9943,9941,9939,9937,9935,9933,9931,9929,9927,9925,9923,9921,9918,9916,9914,9912,9909,9907,9905,
9902,9900,9897,9895,9892,9890,9887,9884,9882,9879,9876,9874,9871,9868,9865,9862,9859,9857,9854,9851,
9848,9845,9841,9838,9835,9832,9829,9826,9822,9819,9816,9812,9809,9806,9802,9799,9795,9792,9788,9785,
9781,9777,9774,9770,9766,9762,9759,9755,9751,9747,9743,9739,9735,9731,9727,9723,9719,9715,9711,9707,
9702,9698,9694,9690,9685,9681,9677,9672,9668,9663,9659,9654,9650,9645,9640,9636,9631,9626,9622,9617,
9612,9607,9602,9598,9593,9588,9583,9578,9573,9568,9563,9557,9552,9547,9542,9537,9531,9526,9521,9515,
9510,9505,9499,9494,9488,9483,9477,9472,9466,9460,9455,9449,9443,9438,9432,9426,9420,9414,9408,9402,
9396,9390,9384,9378,9372,9366,9360,9354,9348,9342,9335,9329,9323,9316,9310,9304,9297,9291,9284,9278,
9271,9265,9258,9252,9245,9238,9232,9225,9218,9211,9205,9198,9191,9184,9177,9170,9163,9156,9149,9142,
9135,9128,9121,9114,9106,9099,9092,9085,9077,9070,9063,9055,9048,9040,9033,9025,9018,9010,9003,8995,
8987,8980,8972,8964,8957,8949,8941,8933,8925,8917,8910,8902,8894,8886,8878,8870,8862,8853,8845,8837,
8829,8821,8813,8804,8796,8788,8779,8771,8763,8754,8746,8737,8729,8720,8712,8703,8694,8686,8677,8668,
8660,8651,8642,8633,8625,8616,8607,8598,8589,8580,8571,8562,8553,8544,8535,8526,8517,8508,8498,8489,
8480,8471,8461,8452,8443,8433,8424,8415,8405,8396,8386,8377,8367,8358,8348,8338,8329,8319,8309,8300,
8290,8280,8270,8260,8251,8241,8231,8221,8211,8201,8191,8181,8171,8161,8151,8141,8131,8120,8110,8100,
8090,8079,8069,8059,8048,8038,8028,8017,8007,7996,7986,7975,7965,7954,7944,7933,7922,7912,7901,7890,
7880,7869,7858,7847,7836,7826,7815,7804,7793,7782,7771,7760,7749,7738,7727,7716,7705,7693,7682,7671,
7660,7649,7637,7626,7615,7604,7592,7581,7569,7558,7547,7535,7524,7512,7501,7489,7477,7466,7454,7443,
7431,7419,7408,7396,7384,7372,7360,7349,7337,7325,7313,7301,7289,7277,7265,7253,7241,7229,7217,7205,
7193,7181,7169,7156,7144,7132,7120,7107,7095,7083,7071,7058,7046,7033,7021,7009,6996,6984,6971,6959,
6946,6934,6921,6908,6896,6883,6870,6858,6845,6832,6819,6807,6794,6781,6768,6755,6743,6730,6717,6704,
6691,6678,6665,6652,6639,6626,6613,6600,6586,6573,6560,6547,6534,6520,6507,6494,6481,6467,6454,6441,
6427,6414,6401,6387,6374,6360,6347,6333,6320,6306,6293,6279,6266,6252,6238,6225,6211,6197,6184,6170,
6156,6142,6129,6115,6101,6087,6073,6059,6045,6032,6018,6004,5990,5976,5962,5948,5934,5920,5906,5891,
5877,5863,5849,5835,5821,5807,5792,5778,5764,5750,5735,5721,5707,5692,5678,5664,5649,5635,5620,5606,
5591,5577,5562,5548,5533,5519,5504,5490,5475,5461,5446,5431,5417,5402,5387,5372,5358,5343,5328,5313,
5299,5284,5269,5254,5239,5224,5210,5195,5180,5165,5150,5135,5120,5105,5090,5075,5060,5045,5030,5015,
4999,4984,4969,4954,4939,4924,4909,4893,4878,4863,4848,4832,4817,4802,4786,4771,4756,4740,4725,4710,
4694,4679,4663,4648,4632,4617,4601,4586,4570,4555,4539,4524,4508,4493,4477,4461,4446,4430,4415,4399,
4383,4368,4352,4336,4320,4305,4289,4273,4257,4241,4226,4210,4194,4178,4162,4146,4131,4115,4099,4083,
4067,4051,4035,4019,4003,3987,3971,3955,3939,3923,3907,3891,3875,3859,3842,3826,3810,3794,3778,3762,
3746,3729,3713,3697,3681,3665,3648,3632,3616,3599,3583,3567,3551,3534,3518,3502,3485,3469,3452,3436,
3420,3403,3387,3370,3354,3338,3321,3305,3288,3272,3255,3239,3222,3206,3189,3173,3156,3139,3123,3106,
3090,3073,3056,3040,3023,3007,2990,2973,2957,2940,2923,2907,2890,2873,2856,2840,2823,2806,2789,2773,
2756,2739,2722,2706,2689,2672,2655,2638,2621,2605,2588,2571,2554,2537,2520,2503,2486,2469,2453,2436,
2419,2402,2385,2368,2351,2334,2317,2300,2283,2266,2249,2232,2215,2198,2181,2164,2147,2130,2113,2096,
2079,2062,2044,2027,2010,1993,1976,1959,1942,1925,1908,1890,1873,1856,1839,1822,1805,1788,1770,1753,
1736,1719,1702,1684,1667,1650,1633,1616,1598,1581,1564,1547,1529,1512,1495,1478,1460,1443,1426,1409,
1391,1374,1357,1339,1322,1305,1287,1270,1253,1236,1218,1201,1184,1166,1149,1132,1114,1097,1079,1062,
1045,1027,1010,993,975,958,941,923,906,888,871,854,836,819,801,784,767,749,732,714,
697,680,662,645,627,610,593,575,558,540,523,505,488,471,453,436,418,401,383,366,
348,331,314,296,279,261,244,226,209,191,174,157,139,122,104,87,69,52,34,17,
0,-17,-34,-52,-69,-87,-104,-122,-139,-157,-174,-191,-209,-226,-244,-261,-279,-296,-314,-331,
-348,-366,-383,-401,-418,-436,-453,-471,-488,-505,-523,-540,-558,-575,-593,-610,-627,-645,-662,-680,
-697,-714,-732,-749,-767,-784,-801,-819,-836,-854,-871,-888,-906,-923,-941,-958,-975,-993,-1010,-1027,
-1045,-1062,-1079,-1097,-1114,-1132,-1149,-1166,-1184,-1201,-1218,-1236,-1253,-1270,-1287,-1305,-1322,-1339,-1357,-1374,
-1391,-1409,-1426,-1443,-1460,-1478,-1495,-1512,-1529,-1547,-1564,-1581,-1598,-1616,-1633,-1650,-1667,-1684,-1702,-1719,
-1736,-1753,-1770,-1788,-1805,-1822,-1839,-1856,-1873,-1890,-1908,-1925,-1942,-1959,-1976,-1993,-2010,-2027,-2044,-2062,
-2079,-2096,-2113,-2130,-2147,-2164,-2181,-2198,-2215,-2232,-2249,-2266,-2283,-2300,-2317,-2334,-2351,-2368,-2385,-2402,
-2419,-2436,-2453,-2469,-2486,-2503,-2520,-2537,-2554,-2571,-2588,-2605,-2621,-2638,-2655,-2672,-2689,-2706,-2722,-2739,
-2756,-2773,-2789,-2806,-2823,-2840,-2856,-2873,-2890,-2907,-2923,-2940,-2957,-2973,-2990,-3007,-3023,-3040,-3056,-3073,
-3090,-3106,-3123,-3139,-3156,-3173,-3189,-3206,-3222,-3239,-3255,-3272,-3288,-3305,-3321,-3338,-3354,-3370,-3387,-3403,
-3420,-3436,-3452,-3469,-3485,-3502,-3518,-3534,-3551,-3567,-3583,-3599,-3616,-3632,-3648,-3665,-3681,-3697,-3713,-3729,
-3746,-3762,-3778,-3794,-3810,-3826,-3842,-3859,-3875,-3891,-3907,-3923,-3939,-3955,-3971,-3987,-4003,-4019,-4035,-4051,
-4067,-4083,-4099,-4115,-4131,-4146,-4162,-4178,-4194,-4210,-4226,-4241,-4257,-4273,-4289,-4305,-4320,-4336,-4352,-4368,
-4383,-4399,-4415,-4430,-4446,-4461,-4477,-4493,-4508,-4524,-4539,-4555,-4570,-4586,-4601,-4617,-4632,-4648,-4663,-4679,
-4694,-4710,-4725,-4740,-4756,-4771,-4786,-4802,-4817,-4832,-4848,-4863,-4878,-4893,-4909,-4924,-4939,-4954,-4969,-4984,
-4999,-5015,-5030,-5045,-5060,-5075,-5090,-5105,-5120,-5135,-5150,-5165,-5180,-5195,-5210,-5224,-5239,-5254,-5269,-5284,
-5299,-5313,-5328,-5343,-5358,-5372,-5387,-5402,-5417,-5431,-5446,-5461,-5475,-5490,-5504,-5519,-5533,-5548,-5562,-5577,
-5591,-5606,-5620,-5635,-5649,-5664,-5678,-5692,-5707,-5721,-5735,-5750,-5764,-5778,-5792,-5807,-5821,-5835,-5849,-5863,
-5877,-5891,-5906,-5920,-5934,-5948,-5962,-5976,-5990,-6004,-6018,-6032,-6045,-6059,-6073,-6087,-6101,-6115,-6129,-6142,
-6156,-6170,-6184,-6197,-6211,-6225,-6238,-6252,-6266,-6279,-6293,-6306,-6320,-6333,-6347,-6360,-6374,-6387,-6401,-6414,
-6427,-6441,-6454,-6467,-6481,-6494,-6507,-6520,-6534,-6547,-6560,-6573,-6586,-6600,-6613,-6626,-6639,-6652,-6665,-6678,
-6691,-6704,-6717,-6730,-6743,-6755,-6768,-6781,-6794,-6807,-6819,-6832,-6845,-6858,-6870,-6883,-6896,-6908,-6921,-6934,
-6946,-6959,-6971,-6984,-6996,-7009,-7021,-7033,-7046,-7058,-7071,-7083,-7095,-7107,-7120,-7132,-7144,-7156,-7169,-7181,
-7193,-7205,-7217,-7229,-7241,-7253,-7265,-7277,-7289,-7301,-7313,-7325,-7337,-7349,-7360,-7372,-7384,-7396,-7408,-7419,
-7431,-7443,-7454,-7466,-7477,-7489,-7501,-7512,-7524,-7535,-7547,-7558,-7569,-7581,-7592,-7604,-7615,-7626,-7637,-7649,
-7660,-7671,-7682,-7693,-7705,-7716,-7727,-7738,-7749,-7760,-7771,-7782,-7793,-7804,-7815,-7826,-7836,-7847,-7858,-7869,
-7880,-7890,-7901,-7912,-7922,-7933,-7944,-7954,-7965,-7975,-7986,-7996,-8007,-8017,-8028,-8038,-8048,-8059,-8069,-8079,
-8090,-8100,-8110,-8120,-8131,-8141,-8151,-8161,-8171,-8181,-8191,-8201,-8211,-8221,-8231,-8241,-8251,-8260,-8270,-8280,
-8290,-8300,-8309,-8319,-8329,-8338,-8348,-8358,-8367,-8377,-8386,-8396,-8405,-8415,-8424,-8433,-8443,-8452,-8461,-8471,
-8480,-8489,-8498,-8508,-8517,-8526,-8535,-8544,-8553,-8562,-8571,-8580,-8589,-8598,-8607,-8616,-8625,-8633,-8642,-8651,
-8660,-8668,-8677,-8686,-8694,-8703,-8712,-8720,-8729,-8737,-8746,-8754,-8763,-8771,-8779,-8788,-8796,-8804,-8813,-8821,
-8829,-8837,-8845,-8853,-8862,-8870,-8878,-8886,-8894,-8902,-8910,-8917,-8925,-8933,-8941,-8949,-8957,-8964,-8972,-8980,
-8987,-8995,-9003,-9010,-9018,-9025,-9033,-9040,-9048,-9055,-9063,-9070,-9077,-9085,-9092,-9099,-9106,-9114,-9121,-9128,
-9135,-9142,-9149,-9156,-9163,-9170,-9177,-9184,-9191,-9198,-9205,-9211,-9218,-9225,-9232,-9238,-9245,-9252,-9258,-9265,
-9271,-9278,-9284,-9291,-9297,-9304,-9310,-9316,-9323,-9329,-9335,-9342,-9348,-9354,-9360,-9366,-9372,-9378,-9384,-9390,
-9396,-9402,-9408,-9414,-9420,-9426,-9432,-9438,-9443,-9449,-9455,-9460,-9466,-9472,-9477,-9483,-9488,-9494,-9499,-9505,
-9510,-9515,-9521,-9526,-9531,-9537,-9542,-9547,-9552,-9557,-9563,-9568,-9573,-9578,-9583,-9588,-9593,-9598,-9602,-9607,
-9612,-9617,-9622,-9626,-9631,-9636,-9640,-9645,-9650,-9654,-9659,-9663,-9668,-9672,-9677,-9681,-9685,-9690,-9694,-9698,
-9702,-9707,-9711,-9715,-9719,-9723,-9727,-9731,-9735,-9739,-9743,-9747,-9751,-9755,-9759,-9762,-9766,-9770,-9774,-9777,
-9781,-9785,-9788,-9792,-9795,-9799,-9802,-9806,-9809,-9812,-9816,-9819,-9822,-9826,-9829,-9832,-9835,-9838,-9841,-9845,
-9848,-9851,-9854,-9857,-9859,-9862,-9865,-9868,-9871,-9874,-9876,-9879,-9882,-9884,-9887,-9890,-9892,-9895,-9897,-9900,
-9902,-9905,-9907,-9909,-9912,-9914,-9916,-9918,-9921,-9923,-9925,-9927,-9929,-9931,-9933,-9935,-9937,-9939,-9941,-9943,
-9945,-9947,-9948,-9950,-9952,-9953,-9955,-9957,-9958,-9960,-9961,-9963,-9964,-9966,-9967,-9969,-9970,-9971,-9973,-9974,
-9975,-9976,-9978,-9979,-9980,-9981,-9982,-9983,-9984,-9985,-9986,-9987,-9988,-9988,-9989,-9990,-9991,-9991,-9992,-9993,
-9993,-9994,-9995,-9995,-9996,-9996,-9997,-9997,-9997,-9998,-9998,-9998,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9999,
-10000,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9999,-9998,-9998,-9998,-9997,-9997,-9997,-9996,-9996,-9995,-9995,-9994,
-9993,-9993,-9992,-9991,-9991,-9990,-9989,-9988,-9988,-9987,-9986,-9985,-9984,-9983,-9982,-9981,-9980,-9979,-9978,-9976,
-9975,-9974,-9973,-9971,-9970,-9969,-9967,-9966,-9964,-9963,-9961,-9960,-9958,-9957,-9955,-9953,-9952,-9950,-9948,-9947,
-9945,-9943,-9941,-9939,-9937,-9935,-9933,-9931,-9929,-9927,-9925,-9923,-9921,-9918,-9916,-9914,-9912,-9909,-9907,-9905,
-9902,-9900,-9897,-9895,-9892,-9890,-9887,-9884,-9882,-9879,-9876,-9874,-9871,-9868,-9865,-9862,-9859,-9857,-9854,-9851,
-9848,-9845,-9841,-9838,-9835,-9832,-9829,-9826,-9822,-9819,-9816,-9812,-9809,-9806,-9802,-9799,-9795,-9792,-9788,-9785,
-9781,-9777,-9774,-9770,-9766,-9762,-9759,-9755,-9751,-9747,-9743,-9739,-9735,-9731,-9727,-9723,-9719,-9715,-9711,-9707,
-9702,-9698,-9694,-9690,-9685,-9681,-9677,-9672,-9668,-9663,-9659,-9654,-9650,-9645,-9640,-9636,-9631,-9626,-9622,-9617,
-9612,-9607,-9602,-9598,-9593,-9588,-9583,-9578,-9573,-9568,-9563,-9557,-9552,-9547,-9542,-9537,-9531,-9526,-9521,-9515,
-9510,-9505,-9499,-9494,-9488,-9483,-9477,-9472,-9466,-9460,-9455,-9449,-9443,-9438,-9432,-9426,-9420,-9414,-9408,-9402,
-9396,-9390,-9384,-9378,-9372,-9366,-9360,-9354,-9348,-9342,-9335,-9329,-9323,-9316,-9310,-9304,-9297,-9291,-9284,-9278,
-9271,-9265,-9258,-9252,-9245,-9238,-9232,-9225,-9218,-9211,-9205,-9198,-9191,-9184,-9177,-9170,-9163,-9156,-9149,-9142,
-9135,-9128,-9121,-9114,-9106,-9099,-9092,-9085,-9077,-9070,-9063,-9055,-9048,-9040,-9033,-9025,-9018,-9010,-9003,-8995,
-8987,-8980,-8972,-8964,-8957,-8949,-8941,-8933,-8925,-8917,-8910,-8902,-8894,-8886,-8878,-8870,-8862,-8853,-8845,-8837,
-8829,-8821,-8813,-8804,-8796,-8788,-8779,-8771,-8763,-8754,-8746,-8737,-8729,-8720,-8712,-8703,-8694,-8686,-8677,-8668,
-8660,-8651,-8642,-8633,-8625,-8616,-8607,-8598,-8589,-8580,-8571,-8562,-8553,-8544,-8535,-8526,-8517,-8508,-8498,-8489,
-8480,-8471,-8461,-8452,-8443,-8433,-8424,-8415,-8405,-8396,-8386,-8377,-8367,-8358,-8348,-8338,-8329,-8319,-8309,-8300,
-8290,-8280,-8270,-8260,-8251,-8241,-8231,-8221,-8211,-8201,-8191,-8181,-8171,-8161,-8151,-8141,-8131,-8120,-8110,-8100,
-8090,-8079,-8069,-8059,-8048,-8038,-8028,-8017,-8007,-7996,-7986,-7975,-7965,-7954,-7944,-7933,-7922,-7912,-7901,-7890,
-7880,-7869,-7858,-7847,-7836,-7826,-7815,-7804,-7793,-7782,-7771,-7760,-7749,-7738,-7727,-7716,-7705,-7693,-7682,-7671,
-7660,-7649,-7637,-7626,-7615,-7604,-7592,-7581,-7569,-7558,-7547,-7535,-7524,-7512,-7501,-7489,-7477,-7466,-7454,-7443,
-7431,-7419,-7408,-7396,-7384,-7372,-7360,-7349,-7337,-7325,-7313,-7301,-7289,-7277,-7265,-7253,-7241,-7229,-7217,-7205,
-7193,-7181,-7169,-7156,-7144,-7132,-7120,-7107,-7095,-7083,-7071,-7058,-7046,-7033,-7021,-7009,-6996,-6984,-6971,-6959,
-6946,-6934,-6921,-6908,-6896,-6883,-6870,-6858,-6845,-6832,-6819,-6807,-6794,-6781,-6768,-6755,-6743,-6730,-6717,-6704,
-6691,-6678,-6665,-6652,-6639,-6626,-6613,-6600,-6586,-6573,-6560,-6547,-6534,-6520,-6507,-6494,-6481,-6467,-6454,-6441,
-6427,-6414,-6401,-6387,-6374,-6360,-6347,-6333,-6320,-6306,-6293,-6279,-6266,-6252,-6238,-6225,-6211,-6197,-6184,-6170,
-6156,-6142,-6129,-6115,-6101,-6087,-6073,-6059,-6045,-6032,-6018,-6004,-5990,-5976,-5962,-5948,-5934,-5920,-5906,-5891,
-5877,-5863,-5849,-5835,-5821,-5807,-5792,-5778,-5764,-5750,-5735,-5721,-5707,-5692,-5678,-5664,-5649,-5635,-5620,-5606,
-5591,-5577,-5562,-5548,-5533,-5519,-5504,-5490,-5475,-5461,-5446,-5431,-5417,-5402,-5387,-5372,-5358,-5343,-5328,-5313,
-5299,-5284,-5269,-5254,-5239,-5224,-5210,-5195,-5180,-5165,-5150,-5135,-5120,-5105,-5090,-5075,-5060,-5045,-5030,-5015,
-5000,-4984,-4969,-4954,-4939,-4924,-4909,-4893,-4878,-4863,-4848,-4832,-4817,-4802,-4786,-4771,-4756,-4740,-4725,-4710,
-4694,-4679,-4663,-4648,-4632,-4617,-4601,-4586,-4570,-4555,-4539,-4524,-4508,-4493,-4477,-4461,-4446,-4430,-4415,-4399,
-4383,-4368,-4352,-4336,-4320,-4305,-4289,-4273,-4257,-4241,-4226,-4210,-4194,-4178,-4162,-4146,-4131,-4115,-4099,-4083,
-4067,-4051,-4035,-4019,-4003,-3987,-3971,-3955,-3939,-3923,-3907,-3891,-3875,-3859,-3842,-3826,-3810,-3794,-3778,-3762,
-3746,-3729,-3713,-3697,-3681,-3665,-3648,-3632,-3616,-3599,-3583,-3567,-3551,-3534,-3518,-3502,-3485,-3469,-3452,-3436,
-3420,-3403,-3387,-3370,-3354,-3338,-3321,-3305,-3288,-3272,-3255,-3239,-3222,-3206,-3189,-3173,-3156,-3139,-3123,-3106,
-3090,-3073,-3056,-3040,-3023,-3007,-2990,-2973,-2957,-2940,-2923,-2907,-2890,-2873,-2856,-2840,-2823,-2806,-2789,-2773,
-2756,-2739,-2722,-2706,-2689,-2672,-2655,-2638,-2621,-2605,-2588,-2571,-2554,-2537,-2520,-2503,-2486,-2469,-2453,-2436,
-2419,-2402,-2385,-2368,-2351,-2334,-2317,-2300,-2283,-2266,-2249,-2232,-2215,-2198,-2181,-2164,-2147,-2130,-2113,-2096,
-2079,-2062,-2044,-2027,-2010,-1993,-1976,-1959,-1942,-1925,-1908,-1890,-1873,-1856,-1839,-1822,-1805,-1788,-1770,-1753,
-1736,-1719,-1702,-1684,-1667,-1650,-1633,-1616,-1598,-1581,-1564,-1547,-1529,-1512,-1495,-1478,-1460,-1443,-1426,-1409,
-1391,-1374,-1357,-1339,-1322,-1305,-1287,-1270,-1253,-1236,-1218,-1201,-1184,-1166,-1149,-1132,-1114,-1097,-1079,-1062,
-1045,-1027,-1010,-993,-975,-958,-941,-923,-906,-888,-871,-854,-836,-819,-801,-784,-767,-749,-732,-714,
-697,-680,-662,-645,-627,-610,-593,-575,-558,-540,-523,-505,-488,-471,-453,-436,-418,-401,-383,-366,
-348,-331,-314,-296,-279,-261,-244,-226,-209,-191,-174,-157,-139,-122,-104,-87,-69,-52,-34,-17
};
/* Result of calculating the matrix multiplications on page 33 of the OV9655_IG_Tomtom 1.0 document. */
/* Note that this is only M1 - M6, as M7 - M9 are fixed in hardware. */
/* formula a * gain * sin( alpha ) + b * gain * cos( alpha ) + c */
static struct color_correct	correct_array[]=
{
	{-16637,	-8968,		0,	0x4F},
	{16635,		-7791,		0,	0x50},
	{61,		21038,		0,	0x51},
	{-8968,		16637,		0,	0x52},
	{-7791,		-16635,		0,	0x53},
	{21038,		-61,		0,	0x54}
};
	
/* Correct the colors. Hue is set by alpha (-360 ~ 360), saturation by gain (0-255, or more to be more precise, (0-255) << 8) */
static int ov9xx0_do_color_correct( struct i2c_client *c, signed long int alpha, signed long int gain )
{
	int			index=0;
	signed long int		result=0;
	unsigned char		signreg=0;
	int			rc=0;

	/* Range 0 - 3600 */
	while( alpha < 0 ) alpha+=3600;
	alpha%=3600;

	/* Range -65 - 65 */
	if( (gain > 66) || (gain < -66) ) gain%=66;

	/* Calculate for each register in the matrix. */
	for( index=0; index < (sizeof( correct_array )/sizeof( correct_array[0] )); index++ )
	{
		/* Calculate the result. Note that the result should be divided by 10000. */
		result=((( (((__s32) correct_array[index].a_10k) * ((__s32) sin10k_array[alpha]) ) / 10000) +
		         ( (((__s32) correct_array[index].b_10k) * ((__s32) cos10k_array[alpha]) ) / 10000) ) * 
		         ((__s32) gain)) + correct_array[index].c_10k;
		result/=10000;

	result, correct_array[index].a_10k, correct_array[index].b_10k, correct_array[index].c_10k, alpha,
	sin10k_array[alpha], cos10k_array[alpha], gain );
		/* If negative we need to set the signbit. */
		if( result & 0x00000100 )
			signreg|=0x00000001 << index;

		/* Program the result in the register. */
		rc=ov_write( c, correct_array[index].target_reg, ((unsigned char) result) );
		if( rc < 0 )
			return rc;
	}

	/* Lastly, program the signreg. */
	rc=ov_write( c, 0x58, signreg );
	if( rc < 0 )
		return rc;

	/* All ok. */
	return 0;
}
#endif
 
static int ov9xx0_det_banding( struct i2c_client *c, unsigned short framerate )
{
	unsigned char	regval=0;
	unsigned char	BandVal50, BandVal60;
	unsigned short	MaxExposLine=0;
	int		rc=0;

	/* Get the current mode. */
	ov_read( c, 0x12, &regval );

	/* This bit determines whether we are in Full res or VGA mode. */
	if( regval & 0x40 )
	{
		/* Full resolution mode. */
		MaxExposLine=1045;
	}
	else
	{
		/* VGA mode. */
		MaxExposLine=495;
	}

	/* Determine 60Hz and 50Hz banding value. */
	BandVal50=framerate*MaxExposLine/100;
	BandVal60=framerate*MaxExposLine/120;
	rc=ov_write( c, 0xA2, BandVal50 );		
	if( rc < 0 )
		return rc;
	rc=ov_write( c, 0xA3, BandVal60 );		
	if( rc < 0 )
		return rc;

	/* Done. */
	return 0;
}

static int ov9xx0_set_control(struct i2c_client *c,
			      struct ovcamchip_control *ctl)
{
	struct ovcamchip	*ov = i2c_get_clientdata(c);
	struct ov9655_private	*priv_dat=ov->spriv;
	int			rc;
	int			v = ctl->value;
	unsigned char		regval=0;

	switch (ctl->id) {
	case OVCAMCHIP_CID_CONT:
	case V4L2_CID_CONTRAST :
		rc = ov_write(c, 0x56, v );
		if( rc < 0 )
			goto out;
		priv_dat->image_properties.contrast=v & 0xFF;
	
		break;
	case OVCAMCHIP_CID_BRIGHT:
	case V4L2_CID_BRIGHTNESS :
		rc = ov_write(c, 0x55, v );
		if( rc < 0 )
			goto out;
		priv_dat->image_properties.brightness=v & 0xFF;
		break;
#ifdef OV_BUILD_HUESAT
	case OVCAMCHIP_CID_SAT  :
	case V4L2_CID_SATURATION:
		rc=ov9xx0_do_color_correct( c, priv_dat->image_properties.hue, ctl->value );
		if( rc < 0 )
			goto out;
		priv_dat->image_properties.saturation=v;
		break;
	case OVCAMCHIP_CID_HUE:
	case V4L2_CID_HUE     :
		rc=ov9xx0_do_color_correct( c, ctl->value, priv_dat->image_properties.saturation );
		if( rc < 0 )
			goto out;
		priv_dat->image_properties.hue=v;
		break;
#endif
	case OVCAMCHIP_CID_EXP:
		/* Set the exposure. 16 bits spread over 3 registers.*/
		rc = ov_write_mask(c, REG_EXP1, v & 0x03, 0x03);
		if( rc < 0 ) goto out;
		rc = ov_write_mask( c, REG_EXP2, (v >> 2) & 0xFF, 0xFF );
		if( rc < 0 ) goto out;
		rc = ov_write_mask(c, REG_EXP3, (v >> 10) & 0x3F, 0x3F);
		if( rc < 0 ) goto out;
		break;
	case OVCAMCHIP_CID_FREQ:
	{
		unsigned char	rgbmod=0;
		unsigned short	rndreg=0;
		unsigned short	actrate=0;

		/* Determine if we're running in max 15fps mode (sxga or 15fps vga) or in 30fps mode (VGA or lower only). */
		if( (ctl->value > 30) || (ctl->value <= 0) )
			return -EINVAL;

		/* Get the current mode. */
		ov9xx0_get_rate( c, &rgbmod, &rndreg );

		/* Now calculate the nearest FPS possible. We don't want to use float, so multiply by 10. */
		actrate=rndreg;
		rndreg/=((unsigned short) v);

		/* Now round to the nearest integer. Also deduct one. CLKRC value is one higher as programmed. */
		regval=((unsigned char) (rndreg/10)) - ((rndreg % 10) >= 5 ? 0 : 1);
		if( regval > (0x3F - rgbmod) ) regval=0x3F;
		else regval+=rgbmod;

		/* Save the framerate. */
		rc=ov_write_mask( c, 0x11, regval, 0x3F );
		if( rc < 0 ) goto out;

		/* Determine the actual framerate, this for the banding value. */
		actrate/=(10*((unsigned short) (regval - rgbmod)) + 10);

		/* Program the banding values. */
		rc=ov9xx0_det_banding( c, actrate );
		if( rc < 0 ) goto out;

		break;
	}

	case OVCAMCHIP_CID_BANDFILT:
		/* Enable/Disable banding filter. */
		rc = ov_write_mask(c, 0x13, v?0x20:0x00, 0x20);
		break;
	case OVCAMCHIP_CID_MIRROR:
		rc = ov_write_mask(c, 0x1e, v?0x20:0x00, 0x20);
		break;
	default:
		PDEBUG(2, "control not supported: %d", ctl->id);
		return -EPERM;
	}

out:
	PDEBUG(3, "id=%d, arg=%d, rc=%d", ctl->id, v, rc);
	return rc;
}

static int ov9xx0_get_control(struct i2c_client *c,
			      struct ovcamchip_control *ctl)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov9655_private	*priv_dat=ov->spriv;
	int rc = 0;
	unsigned char val = 0;

	switch (ctl->id) {
	case OVCAMCHIP_CID_CONT:
	case V4L2_CID_CONTRAST :
		ctl->value=priv_dat->image_properties.contrast;
		break;
	case OVCAMCHIP_CID_BRIGHT:
	case V4L2_CID_BRIGHTNESS :
		ctl->value=priv_dat->image_properties.brightness;
		break;
#ifdef OV_BUILD_HUESAT
	case OVCAMCHIP_CID_SAT:
	case V4L2_CID_SATURATION:
		ctl->value=priv_dat->image_properties.saturation;
		break;
	case OVCAMCHIP_CID_HUE:
	case V4L2_CID_HUE     :
		ctl->value=priv_dat->image_properties.hue;
		break;
#endif
	case OVCAMCHIP_CID_EXP:
		rc = ov_read( c, REG_EXP1, &val );
		if( rc < 0 ) return rc;
		ctl->value=(int) (val & 0x03);
		rc = ov_read(c, REG_EXP2, &val );
		if( rc < 0 ) return rc;
		ctl->value|=((int) val) << 2;
		rc = ov_read(c, REG_EXP3, &val );
		if( rc < 0 ) return rc;
		ctl->value|=((int) (val & 0x3F)) << 10;
		break;
	case OVCAMCHIP_CID_FREQ:
	{
		unsigned char	rgbmod=0;
		unsigned short	rndreg=0;

		/* Get the current mode. */
		ov9xx0_get_rate( c, &rgbmod, &rndreg );

		/* Get the register value. */
		ov_read( c, 0x11, &val );
		val-=rgbmod;

		/* Now calculate the nearest FPS possible. We don't want to use float, so multiply by 10. */
		ctl->value=rndreg/((unsigned short) (val + 1) );
		ctl->value/=10;
		break;
	}
	case OVCAMCHIP_CID_BANDFILT:
		rc = ov_read(c, 0x13, &val);
		if( rc < 0 ) return rc;
		ctl->value=((val & 0x20) ? 0x01:0x00);
		break;
	case OVCAMCHIP_CID_MIRROR:
		rc = ov_read(c, 0x1e, &val );
		if( rc < 0 ) return rc;
		ctl->value=((val & 0x20) ? 0x01 : 0x00);
		break;
	default:
		PDEBUG(2, "control not supported: %d", ctl->id);
		return -EPERM;
	}

	PDEBUG(3, "id=%d, arg=%d, rc=%d", ctl->id, ctl->value, rc);
	return rc;
}

static int ov9xx0_mode_init(struct i2c_client *c, struct ovcamchip_window *win)
{
	/* First check the format. */
	switch( win->format )
	{
		case VIDEO_PALETTE_YUV422:
		case VIDEO_PALETTE_YUV420:
			/* Conversion from YUV422 to YUV420 is done by the 23C2413, so we program the same. */
			ov_write_mask( c, 0x0C, 0x04, 0x04 );		// RGB565 Output data average.
			ov_write_mask( c, 0x12, 0x02, 0x07 );		// Select YUV output format. bit[2] low for YUV format.
			ov_write_mask( c, 0x3A, 0x80, 0xCC );		// From the specs. 7:6=10 PCLK 4ns delay, 3:2=00 (YUYV)
			ov_write_mask( c, 0x40, 0x00, 0x30 );		// Select normal RGB format.
                        break;

		case VIDEO_PALETTE_RAW:
			/* Assuming this is RAW RGB. */
			ov_write_mask( c, 0x0C, 0x04, 0x04 );		// RGB565 Output data average.
			ov_write_mask( c, 0x12, 0x04, 0x07 );		// Raw RGB output format. bit[2] high for RGB format.
			ov_write_mask( c, 0x3A, 0xC0, 0xCC );		// From the specs. 7:6=11 PCLK 6ns delay, 3:2=00 (YUYV)
			ov_write_mask( c, 0x40, 0x00, 0x30 );		// Select normal RGB format.
                        break;

		case VIDEO_PALETTE_RGB565:
			/* RGB 565 output format. */
			ov_write_mask( c, 0x0C, 0x00, 0x04 );		// RGB565 RGB565 format.
			ov_write_mask( c, 0x12, 0x07, 0x07 );		// Select RGB output format. bit[2] high for RGB format.
			ov_write_mask( c, 0x3A, 0xC0, 0xCC );		// From the specs. 7:6=11 PCLK 6ns delay, 3:2=00 (YUYV)
			ov_write_mask( c, 0x40, 0x10, 0x30 );		// Select RGB565 format.
			break;

		case VIDEO_PALETTE_RGB555:
			/* RGB 555 output format. */
			ov_write_mask( c, 0x0C, 0x04, 0x04 );		// RGB565 Output data average.
			ov_write_mask( c, 0x12, 0x07, 0x07 );		// Select RGB output format. bit[2] high for RGB format.
			ov_write_mask( c, 0x3A, 0xC0, 0xCC );		// From the specs. 7:6=11 PCLK 6ns delay, 3:2=00 (YUYV)
			ov_write_mask( c, 0x40, 0x30, 0x30 );		// Select RGB555 format.
			break;

		default :
			/* Unsupported video palette. */
			return -1;
	}

	/* Now the resolution. */
	switch( win->width )
	{
		// Can't test. Not enough coherent memory.
		case 1280 :
			if( win->height == 1024 )
			{
				/* 1280x1024@15fps. SXGA*/
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0xF8 );		// Array Reference Control (AREF3)=0xF8
				ov_write_mask( c, 0x69, 0x02, 0x0E );	// No VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					ov_write_mask( c, 0x11, 0x01, 0x3F );
									// Half pixel clock for RGB raw.
				else
					ov_write_mask( c, 0x11, 0x00, 0x3F );
									// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x00, 0x70 );	// Full resolution/15fps VGA
				ov_write_mask( c, 0x8c, 0x0D, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write( c, 0x41, 0x00 );		// Color matrix coefficient double option=0, Scaling=normal
				ov_write( c, 0x72, 0x00 );		// Vertical pixel:      Use pixel average data,
									//                      output index normal
									// Horizontal pixel:    Use pixel average data,
									//                      output index normal
				ov_write_mask( c, 0x73, 0x01, 0x0F );	// Pixel Clock Divider=2 pixel clock/byte;
				ov_write( c, 0xC7, 0x80 );		// Pixel Clock Frequency Selector=0x80;
				ov_write_mask( c, 0x3E, 0x0C, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=off
				ov_write( c, 0x74, 0x3A );		// Horizontal scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x75, 0x35 );		// Vertical scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=1

				ov_write( c, 0x03, 0x1B );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0xFF );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0xBD );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x1D );		// Horizontal res end=0x280 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1A, 0x81 );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xE2 );
				ov_write_mask( c, 0xA1, 0x01, 0x1F );	// AEC[15:10]=0x01 Maximum Exposure Line Limitation (1045). 
				ov_write( c, 0x10, 0x05 );		// AEC[9:2]=0x05 
				ov_write_mask( c, 0x04, 0x01, 0x03 );	// AEC[15:10]=0x01 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 15 );		// Maximum rate. Set to 15 FPS.
				break;
			}
			else return -2;
			break;

		case 640 :
			// Shifting wrap around on right side. Colors are very wrong. Too dark/red
			if( win->height == 480 )
			{
				/* 640x480@30fps VGA. */
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0xFA );		// Array Reference Control (AREF3)=0xFA
				ov_write_mask( c, 0x69, 0x0A, 0x0E );	// Use VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					ov_write_mask( c, 0x11, 0x03, 0x3F );
									// Half pixel clock for RGB raw.
				else
					ov_write_mask( c, 0x11, 0x01, 0x3F );
									// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x60, 0x70 );	// 30fps VGA with vario pixel.
				ov_write_mask( c, 0x8c, 0x8D, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write_mask( c, 0x41, 0x00, 0x03 );	// Color matrix coefficient double option=0, Scaling=normal
				ov_write( c, 0x72, 0x00 );		// Vertical pixel:      Use pixel average data,
									//                      output index normal
									// Horizontal pixel:    Use pixel average data,
									//                      output index normal
				ov_write_mask( c, 0x73, 0x00, 0x0F );	// Pixel Clock Divider=1 pixel clock/byte;
				ov_write( c, 0xC7, 0x80 );		// Pixel Clock Frequency Selector=0x80;
				ov_write_mask( c, 0x3E, 0x0C, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=off
				ov_write( c, 0x74, 0x3A );		// Horizontal scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x75, 0x35 );		// Vertical scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=0

				ov_write( c, 0x03, 0x12 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0xFF );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0x02 );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x16 );		// Horizontal res end=0x280 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1a, 0x3D );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xAA );
				ov_write_mask( c, 0xA1, 0x00, 0x1F );	// AEC[15:10]=0x00 Maximum Exposure Line Limitation (495). 
				ov_write( c, 0x10, 0x7B );		// AEC[9:2]=0x7B 
				ov_write_mask( c, 0x04, 0x03, 0x03 );	// AEC[15:10]=0x03 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 30 );		// Maximum rate. Set to 30 FPS.
				break;
			}
			else return -2;
			break;
	
		case 320 :
			// Shifting wrap around on right side.
			if( win->height == 240 )
			{
				/* 320x240@30fps QVGA. */
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0x3A );		// Array Reference Control (AREF3)=0xFA
				ov_write_mask( c, 0x69, 0x0A, 0x0E );	// Use VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					return -2;			// Not supported.
				ov_write_mask( c, 0x11, 0x01, 0x3F );	// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x60, 0x70 );	// 30fps VGA with vario pixel.
				ov_write_mask( c, 0x8c, 0x80, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write( c, 0x41, 0x01 );		// Color matrix coefficient double option=0, Scaling=down
				ov_write( c, 0x72, 0x11 );		// Vertical pixel:      Vertical truncate
									//			Drop unused pixel data,
									//                      output index one for every 2 
									// Horizontal pixel:    Horizontal truncate
									//			Drop unused pixel data,
									//                      output index one for every 2 
				ov_write_mask( c, 0x73, 0x01, 0x0F );	// Pixel Clock Divider=2 pixel clock/byte
				ov_write( c, 0xC7, 0x81 );		// Pixel Clock Frequency Selector=0x01;
				ov_write_mask( c, 0x3E, 0x0E, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=off
				ov_write( c, 0x74, 0x10 );		// Horizontal scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x75, 0x10 );		// Vertical scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=1

				ov_write( c, 0x03, 0x02 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0x24 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0x04 );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x18 );		// Horizontal res end=0x140 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1a, 0x81 );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xAA );
				ov_write_mask( c, 0xA1, 0x00, 0x1F );	// AEC[15:10]=0x00 Maximum Exposure Line Limitation (495). 
				ov_write( c, 0x10, 0x7B );		// AEC[9:2]=0x7B 
				ov_write_mask( c, 0x04, 0x03, 0x03 );	// AEC[15:10]=0x03 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 30 );		// Maximum rate. Set to 30 FPS.
				break;
			}
			else return -2;
			break;
	
		case 160 :
			if( win->height == 120 )
			{
				/* 160x120@30fps QQVGA. */
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0x3A );		// Array Reference Control (AREF3)=0xFA
				ov_write_mask( c, 0x69, 0x0A, 0x0E );	// Use VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					return -2;			// Not supported.
				ov_write_mask( c, 0x11, 0x01, 0x3F );	// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x60, 0x70 );	// 30fps VGA with vario pixel.
				ov_write_mask( c, 0x8c, 0x80, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write( c, 0x41, 0x01 );		// Color matrix coefficient double option=0, Scaling=down
				ov_write( c, 0x72, 0x22 );		// Vertical pixel:      Vertical truncate
									//			Drop unused pixel data,
									//                      output index one for every 4 
									// Horizontal pixel:    Horizontal truncate
									//			Drop unused pixel data,
									//                      output index one for every 4 
				ov_write_mask( c, 0x73, 0x02, 0x0F );	// Pixel Clock Divider=4 pixel clock/byte;
				ov_write( c, 0xC7, 0x82 );		// Pixel Clock Frequency Selector=0x82;
				ov_write_mask( c, 0x3E, 0x0E, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=off
				ov_write( c, 0x74, 0x10 );		// Horizontal scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x75, 0x10 );		// Vertical scale down coefficient=0x20 (no digital zoom).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=1

				ov_write( c, 0x03, 0x02 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0xA4 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0x04 );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x18 );		// Horizontal res end=0x140 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1a, 0x81 );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xAA );
				ov_write_mask( c, 0xA1, 0x00, 0x1F );	// AEC[15:10]=0x00 Maximum Exposure Line Limitation (495). 
				ov_write( c, 0x10, 0x7B );		// AEC[9:2]=0x7B 
				ov_write_mask( c, 0x04, 0x03, 0x03 );	// AEC[15:10]=0x03 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 30 );		// Maximum rate. Set to 30 FPS.
				break;
			}
			else return -2;
			break;
	
		case 352 :
			// Shifting wrap around on right side.
			if( win->height == 288 )
			{
				/* 352x288@30FPS CIF */
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0x3A );		// Array Reference Control (AREF3)=0xFA
				ov_write_mask( c, 0x69, 0x0A, 0x0E );	// Use VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					ov_write_mask( c, 0x11, 0x03, 0x3F );
									// Half pixel clock for RGB raw.
				else
					ov_write_mask( c, 0x11, 0x01, 0x3F );
									// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x60, 0x70 );	// 30fps VGA with vario pixel.
				ov_write_mask( c, 0x8c, 0x80, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write( c, 0x41, 0x01 );		// Color matrix coefficient double option=0, Scaling=normal
				ov_write( c, 0x72, 0x00 );		// Vertical pixel:      Use pixel average data,
									//                      output index normal
									// Horizontal pixel:    Use pixel average data,
									//                      output index normal
				ov_write_mask( c, 0x73, 0x01, 0x0F );	// Pixel Clock Divider=1 pixel clock/byte;
				ov_write( c, 0xC7, 0x81 );		// Pixel Clock Frequency Selector=0x81;
				ov_write_mask( c, 0x3E, 0x0E, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=on
				ov_write( c, 0x74, 0x3A );		// Horizontal scale down coefficient=0x3A (zoom to 352).
				ov_write( c, 0x75, 0x35 );		// Vertical scale down coefficient=0x35 (zoom to 288).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=1

				ov_write( c, 0x03, 0x09 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0xE4 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0x0C );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x18 );		// Horizontal res end=0x140 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1a, 0x81 );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xAA );
				ov_write_mask( c, 0xA1, 0x00, 0x1F );	// AEC[15:10]=0x00 Maximum Exposure Line Limitation (495). 
				ov_write( c, 0x10, 0x7B );		// AEC[9:2]=0x7B 
				ov_write_mask( c, 0x04, 0x03, 0x03 );	// AEC[15:10]=0x03 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 30 );		// Maximum rate. Set to 30 FPS.
				break;
			}
			else return -2;
			break;
	
		case 176 :
			if( win->height == 144 )
			{
				/* 176x144@30FPS QCIF */
				/* We assume maximum framerate is needed. If not, can be adjusted later. */
				ov_write( c, 0x36, 0x3A );		// Array Reference Control (AREF3)=0xFA
				ov_write_mask( c, 0x69, 0x0A, 0x0E );	// Use VarioPixel
				if( win->format == VIDEO_PALETTE_RAW )
					ov_write_mask( c, 0x11, 0x03, 0x3F );
									// Half pixel clock for RGB raw.
				else
					ov_write_mask( c, 0x11, 0x01, 0x3F );
									// Maximum frequency else.

				ov_write_mask( c, 0x12, 0x60, 0x70 );	// 30fps VGA with vario pixel.
				ov_write_mask( c, 0x8c, 0x80, 0x8D );	// UV Adjust option=11b, UV average=on
				ov_write_mask( c, 0x6B, 0x40, 0xC0 );	// PLL frequency selection=4x.
				ov_write( c, 0x41, 0x01 );		// Color matrix coefficient double option=0, Scaling=normal
				ov_write( c, 0x72, 0x11 );		// Vertical pixel:      Use pixel average data,
									//                      output index normal
									// Horizontal pixel:    Use pixel average data,
									//                      output index normal
				ov_write_mask( c, 0x73, 0x02, 0x0F );	// Pixel Clock Divider=1 pixel clock/byte;
				ov_write( c, 0xC7, 0x82 );		// Pixel Clock Frequency Selector=0x82;
				ov_write_mask( c, 0x3E, 0x0E, 0x0E );	// Black pixel correction=on, white pixel correction=on,
									// Zoom=on
				ov_write( c, 0x74, 0x3A );		// Horizontal scale down coefficient=0x3A (zoom to 352).
				ov_write( c, 0x75, 0x35 );		// Vertical scale down coefficient=0x35 (zoom to 288).
				ov_write( c, 0x76, 0x01 );		// Pixel Clock Delay=1

				ov_write( c, 0x03, 0x12 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x32, 0xE4 );		// VREF end[2:0]=0, VREF start[2:0]=0.
				ov_write( c, 0x18, 0x0C );		// Horizontal res start=0x00 (shift right by 3)
				ov_write( c, 0x17, 0x18 );		// Horizontal res end=0x140 (shift right by 3)
				ov_write( c, 0x19, 0x01 );		// Vertical res start = 0x14 (shift right by 3)
				ov_write( c, 0x1a, 0x81 );		// Vertical res end = 0x104 (shift right by 3)
				ov_write( c, 0xC0, 0xAA );
				ov_write_mask( c, 0xA1, 0x00, 0x1F );	// AEC[15:10]=0x00 Maximum Exposure Line Limitation (495). 
				ov_write( c, 0x10, 0x7B );		// AEC[9:2]=0x7B 
				ov_write_mask( c, 0x04, 0x03, 0x03 );	// AEC[15:10]=0x03 

				/* Program the banding values. */
				ov9xx0_det_banding( c, 30 );		// Maximum rate. Set to 30 FPS.
				break;
			}
			else return -2;
			break;
	
		default :
			/* Unsupported resolution. */
			return -2;
	}

	/* Done. */
	return 0;
}

static unsigned char ov9xx0_state[0xCF];

static int ov9xx0_command(struct i2c_client *c, unsigned int cmd, void *arg)
{
	switch (cmd) {
	case VIDIOC_S_CTRL:
	case OVCAMCHIP_CMD_S_CTRL:
		return ov9xx0_set_control(c, arg);
	case VIDIOC_G_CTRL:
	case OVCAMCHIP_CMD_G_CTRL:
		return ov9xx0_get_control(c, arg);
	case OVCAMCHIP_CMD_S_MODE:
		return ov9xx0_mode_init(c, arg);
	case OVCAMCHIP_CAM_SLEEP:
		return ov_write_mask( c, 0x39, 0x08, 0x08 );
	case OVCAMCHIP_CAM_WAKE:
		return ov_write_mask( c, 0x39, 0x00, 0x08 );
	case OVCAMCHIP_CAM_SUSPEND:
	{
		int	count;

		/* Save all ov9655 registers. */
		for( count=0; count <= sizeof( ov9xx0_state ); count++ )
			ov_read( c, count, &(ov9xx0_state[count]) );

		/* Set bit 3. This to lower power usage. */
		ov_write_mask( c, 0x39, 0x08, 0x08 );

		/* Tristate clock and output. */
		ov_write_mask( c, 0x0D, 0x06, 0x06 );
		return 0;
	}

	case OVCAMCHIP_CAM_RESUME:
	{
		int		count;
		unsigned char	regval;

		/* Reset the chip */
		ov_write( c, 0x12, 0x80 );
		msleep( 100 );

		/* Restore. */
		for( count=0; count <= sizeof( ov9xx0_state ); count++ )
			ov_write( c, count, ov9xx0_state[count] );
		msleep( 100 );

		/* Dummy read to sync I2C */
		ov_read( c, 0x00, &regval );
		msleep( 100 );
		return 0;
	}

	case OVCAMCHIP_CAM_READREG:
	{
		struct ovcamchip_reg	*reg=(struct ovcamchip_reg *) arg;
		ov_read( c, reg->reg, &(reg->val) );
		reg->val&=reg->mask;
		return 0;
	}

	case OVCAMCHIP_CAM_WRITEREG:
	{
		struct ovcamchip_reg	*reg=(struct ovcamchip_reg *) arg;
		ov_write_mask( c, reg->reg, reg->val, reg->mask );
		return 0;
	}

	default:
		PDEBUG(2, "command not supported: %d", cmd);
		return -ENOIOCTLCMD;
	}
}

struct ovcamchip_ops ov9xx0_ops = {
	.init    =	ov9xx0_init,
	.free    =	ov9xx0_free,
	.command =	ov9xx0_command,
};
